"""
Handling of sets of EnzoHierarchy objects

Largely deprecated.

Author: Matthew Turk <matthewturk@gmail.com>
Affiliation: KIPAC/SLAC/Stanford
Homepage: http://yt.enzotools.org/
License:
  Copyright (C) 2007-2009 Matthew Turk.  All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from yt.lagos import *
import ConfigParser, time, os, sys

class EnzoRun:
    """
    A class that is used to hold the information about an entire Enzo
    Simulation.  This includes all of the datadumps, the parameter file,
    and possibly even the initial conditions.
    """
    def __init__(self, metaData, outputs=None, runFilename=None, classType=EnzoHierarchy, timeID=None, getPFs = True):
        """
        We're going to try to avoid setting too many of the parameters here,
        as many will be changed on and off.  However, we can definitely set all
        of the parameters that are fixed from the initial conditions -- things
        like root grid conditions, problem type, and so on.  While these *may*
        change during the run (for instance, changing to SN problem type) they
        won't do so without quite a bit of effort on the part of the user, and
        we thus don't quite have to care about them too much.

        This is primarily a storage container.  If we open it up, and add to
        it all of our datadumps, we can snag lots of fun information from them
        all.

        @param metaData: text describing the run as a whole.
        @type metaData: string
        @keyword outputs: outputs to add to the run
        @type outputs: list of L{EnzoHierarchies<EnzoHierarchy>}
        """
        if outputs == None: outputs = []
        self.metaData = metaData
        self.classType = classType
        self.outputs = obj.array(outputs)       # Object array of EnzoHierarchies
        self.timesteps = na.zeros(shape=self.outputs.shape, dtype=nT.Float64) # Timesteps
        self.runFilename = runFilename
        self.gotRuns = False
        if runFilename != None and getPFs:
            self.Import(runFilename)
        if not timeID:
            timeID = int(time.time())
        self.timeID = timeID

    def promoteType(self, outputID):
        if hasattr(self.outputs[outputID], "grids"):
            return
        pp = self.outputs[outputID]
        pf_name = os.path.join(pp.fullpath, pp.basename)
        self.outputs[outputID] = EnzoHierarchy(pf_name)
        del pp

    def demoteType(self, outputID):
        if not hasattr(self.outputs[outputID], "grids"):
            return
        pp = self.outputs[outputID]
        pf_name = os.path.join(pp.fullpath, pp.basename)
        self.outputs[outputID] = EnzoParameterFile(pf_name)
        del pp

    def sortOutputs(self):
        """
        Sorts outputs, solely by the time at which they were created.

        Note that this may not be what we want -- we may actually want the time
        in the simulation at which they were dumped.
        """
        order = na.argsort(self.timesteps)
        self.outputs = self.outputs[order]
        self.timesteps = self.timesteps[order]

    def addOutput(self, hierarchy):
        """
        Add an output.  Also, sorts.

        @param hierarchy: either a single hierarchy or a list of hierarchies
        @type hierarchy: L{EnzoHierarchy<EnzoHierarchy>}
        """
        if not isinstance(hierarchy, types.ListType):
            hierarchy = [hierarchy]
        # Our arrays are both one-d, so we'll just extend them
        t = []
        for h in hierarchy:
            t.append(h["InitialTime"])
        self.outputs = obj.array(self.outputs.tolist() + hierarchy)
        self.timesteps=na.array(self.timesteps.tolist() + t,dtype=nT.Float64)
        self.sortOutputs()
        self.gotRuns = True

    def addOutputByFilename(self, filename, hdf_version=4):
        """
        Feed it a list of parameter files, andi t will add 'em all

        @param filename: either a single filename or a list of filenames
        @keyword hdf_version: version of hdf to use
        @type hdf_version: int
        """
        if not isinstance(filename, types.ListType):
            filename = [filename]
        k = []
        for fn in filename:
            mylog.debug("Adding %s to EnzoRun '%s'", fn, self.metaData)
            k.append(self.classType(fn, hdf_version=hdf_version))
            k[-1].run = self
        self.addOutput(k)

    def getCommandLine(self):
        return "./enzo_red_i9_r16"

    def getTime(self):
        return time.ctime(float(self.timeID))


    def runFunction(self, func, args, fmt_string = None):
        """
        Here we can call runFunction, feeding it a function and some arguments.

        The function will be called with every EnzoHierarchy as the first
        argument, and the *args as the rest of the arguments.

        @param func: function handler to be called for every EnzoHierarchy
                     note that it will be called as such:
                     func(EnzoHierarchy, [fmt_string % index], *args)
        @type func: function
        @param args: a list of arguments to pass in
        @type args: list
        @param fmt_string: this is an optional argument, which will be %'d
                           against the index of the parameter file.  Note that
                           it is only %'d against the *index*, so even if the
                           thing is called DataDump1042, if it's the first one, it'll get 0.
        @type fmt_string: string
        """
        if not isinstance(args, types.ListType):
            args = [args]
        for i in range(self.outputs.shape[0]):
            pid = os.fork()
            if pid:
                newpid, exit = os.wait()
                mylog.info("Exit status %s from PID %s", exit, newpid)
            else:
                mylog.debug("Forked process reporting for duty")
                self.promoteType(i)
                a = [self.outputs[i]]
                if fmt_string != None:
                    s = fmt_string % i
                    a += [s]
                a += args
                mylog.debug("Calling %s on %s", func.func_name, a[0].parameterFilename)
                func(*a)
                mylog.debug("Done calling %s, now dying", func.func_name)
                sys.exit()

    def getBefore(self, time):
        return na.where(self.timesteps <= time)[0]

    def getAfter(self, time):
        return na.where(self.timesteps > time)[0]

    def removeOutput(self, key):
        pp = self[key]
        id = self.index(key)
        # Okay, I know this sucks, but, it seems to be the only way, since
        # concatenate doesn't want to work with ObjectArrays
        newOutputs = obj.array(self.outputs[:id].tolist() + self.outputs[id+1:].tolist())
        newTimesteps = obj.array(self.timesteps[:id].tolist() + self.timesteps[id+1:].tolist())
        self.outputs = newOutputs
        self.timesteps = newTimesteps
        del pp, newOutputs, newTimesteps

    def removeOutputFiles(self, outputID):
        """
        This function handles things a bit  more simply than
        removeOutputFilesByGrid, in that it tosses everything over to fido.
        This *deletes* files.  Note that this can be DANGEROUS.  Use with care.

        @param outputID: the ID or basename of the parameter file to toast
        @type outputID: int or string
        """
        import yt.fido
        mylog.info("Deleting %s" % (self[outputID].fullpath))
        #yt.fido.deleteFiles(self.outputs[outputID].fullpath, \
                #os.path.basename(self.outputs[outputID].parameterFilename))
        yt.fido.deleteOutput(self[outputID].fullpath)

    def moveOutput(self, outputID, dest):
        """
        This moves a data directory tree to another location.  This assumes the
        data is already buried.

        @param outputID: the output we want to move
        @type outputID: integer
        @param dest: the basename of the new location
        @type dest: string
        """
        import yt.fido
        pf = self[outputID]
        mylog.info("Moving %s to %s", os.path.basename(pf.fullpath), \
                    dest)
        pf.fullpath = yt.fido.moveOutput(pf.fullpath, os.path.abspath(dest))

    def moveOutputFiles(self, outputID, dest):
        import yt.fido
        pf = self[outputID]
        mylog.info("Moving %s to %s", os.path.basename(pf.fullpath), \
                    dest)
        yt.fido.moveFiles(os.path.abspath(dest), \
            os.path.basename(pf.parameterFilename), \
            extraFiles = [], wd=pf.directory)

    def removeOutputFilesByGrid(self, outputID):
        """
        This function deletes all the files associated with a given parameter
        file, and then the instance itself.  I am of two minds here; do we want
        to execute the costly operation of removing each file associated with
        the hierarchy instance, or do we want to just kill 'em all off via a
        system command?
        
        @param outputID: the ID of the parameter file to toast
        @type outputID: int
        @deprecated: Use removeOutputFiles
        """
        run = self.outputs[outputID]
        for grid in run.grids:
            mylog.debug("Deleting %s", grid.filename)
            os.unlink(grid.filename)
        mylog.debug("Deleting parameter file %s", run.parameterFilename)
        os.unlink(run.parameterFilename)
        mylog.debug("Deleting hierarchy %s", run.hierarchyFilename)
        os.unlink(run.hierarchyFilename)
        mylog.debug("Deleting boundary %s", run.boundaryFilename)
        os.unlink(run.boundaryFilename)
        mylog.debug("Deleting boundary.hdf  %s", run.boundaryFilename+".hdf")
        os.unlink(run.boundaryFilename+".hdf")
        mylog.debug("Deleted everything, I think")

    def Export(self, nn):
        """
        This function returns a ConfigParser object that describes the enough
        that it can be resurrected at a later date.

        @return: ConfigParser
        """
        cfg = ConfigParser.ConfigParser()
        if os.path.exists(nn):
            cfg.read(nn)
        if not cfg.has_section("RunInformation"):
            cfg.add_section("RunInformation")
        cfg.set("RunInformation","MetaData",self.metaData)
        cfg.set("RunInformation","User", os.getenv("USER"))
        cfg.set("RunInformation","ExportTime", time.ctime())
        # Hm, I guess that's all we need for the metadata...
        # Now the question, really, is going to be how do we handle the data
        # listings?
        self.UpdateOutputs(cfg)
        return cfg

    def Import(self, filename):
        """
        Here we read in an EnzoRun file and import the parameter files
        @param filename: the run filename
        @type filename: string
        """
        cc = ConfigParser.ConfigParser()
        cc.read(filename)
        self.metaData = cc.get("RunInformation","MetaData")
        new_outputs = []
        for sec in cc.sections():
            if sec.startswith("Output"):
                # Okay, so it's an output, now what?
                d = cc.get(sec,"Directory")
                f = cc.get(sec,"Basename")
                new_outputs.append(os.path.join(d,f))
        self.gotRuns = True
        self.addOutputByFilename(new_outputs)

    def UpdateOutputs(self, cfg):
        """
        This function accepts a filename, and will update the ConfigParser
        object corresponding to that filename, ensuring that all outputs I{currently}
        associated with the run instance are in the file.  (It does this by
        I{deleting} the [Outputs] section, by the way, and re-initializing it.)
        @param cfg: The filename or ConfigParser object to update
        @type cfg: string or instantiated ConfigParser
        """
        if isinstance(cfg, types.StringType):
            # We open the file, and we will write to it when done
            cc = ConfigParser.ConfigParser()
            cc.read(cfg)
            toWrite = True
        else:
            # Assume it is a ConfigParser object, and let exception get thrown
            # otherwise.
            cc = cfg
            toWrite = False
        for sec in cc.sections():
            if sec.startswith("Output"):
                cc.remove_section(sec)
        num = self.outputs.shape[0]
        otherParams=ytcfg.get("Fido","OtherParamsToStore").split(",")
        for i in range(num):
            n = "Output%04i" % (i)
            cc.add_section(n)
            cc.set(n,"CurrentTimeIdentifier",self.outputs[i]["CurrentTimeIdentifier"])
            cc.set(n,"Directory",self.outputs[i].fullpath)
            cc.set(n,"Basename",self.outputs[i].basename)
            for op in otherParams:
                if self.outputs[i].has_key(op):
                    cc.set(n,op,self.outputs[i][op])
        if toWrite:
            cc.write(cfg)
        # We return *nothing*, because we either modify the ConfigParser in
        # place or we update and write out

    def __getitem__(self, key):
        """
        Based on the type of input, we either return based on index or
        basename.
        """
        if isinstance(key, types.StringType):
            index = self.index(key)
        elif isinstance(key, types.IntType):
            index = key
        return self.outputs[index]

    def index(self, key):
        t = os.path.basename(key)
        # Find out the index
        index = None
        for i in range(self.outputs.shape[0]):
            if self.outputs[i].basename == t:
                index = i
                break
        if index == None:
            raise KeyError
        return index

    def getOutputs(self):
        if not self.gotRuns:
            self.Import(self.runFilename)
        return self.outputs.tolist()

    def __xattrs__(self, mode="default"):
        return("metaData", "getTime()")

    def __repr__(self):
        return self.metaData

    def __len__(self):
        if not self.gotRuns:
            self.Import(self.runFilename)
        return self.outputs.shape[0]

    def __iter__(self):
        if not self.gotRuns:
            self.Import(self.runFilename)
        for i in range(len(self)):
            yield self.outputs[i]

    def keys(self):
        """
        Returns a list of parameterfile basenames
        """
        keys = []
        for i in range(self.outputs.shape[0]):
            keys.append(self.outputs[i].basename)
        return keys



"""
Fido is a module for moving and storing data.  It is designed to watch for
outputs, as well as transparent archiving them during and after a simulation
run.  Additionally, function handlers can be called to deal with output upon
its creation.

Author: Matthew Turk <matthewturk@gmail.com>
Affiliation: KIPAC/SLAC/Stanford
Homepage: http://yt.enzotools.org/
License:
  Copyright (C) 2007-2009 Matthew Turk.  All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from yt.logger import fidoLogger as mylog
from yt.config import ytcfg
from yt.arraytypes import *

import os, os.path, shutil, time, sys, glob, types

OTHERFILES=["rates.out","cool_rates.out"]
NEW_OUTPUT_CREATED = ytcfg.get("fido","NewOutputCreated")
GlobPatterns = ytcfg.get("fido","GlobPatterns").split(",")
NewDirectoryPattern = ytcfg.get("fido","NewDirectoryPattern",raw=True)

# Let's define some useful functions.  Maybe these should go elsewhere?

def get_parameter_line(filename, parameter):
    f = open(filename)
    lines = filter(lambda a: a.startswith(parameter), f)
    if len(lines) == 0:
        raise KeyError(parameter)
    elif len(lines) > 1:
        raise KeyError, "More than one line matches that parameter!"
    else:
        return lines[0]

def get_parent_dir(filename):
    return os.path.normpath( \
            os.path.split(os.path.dirname(os.path.abspath(filename)))[0])

from OutputCollection import *
from FileHandling import *
from OutputWatcher import *
from RunStandalones import *
from ParameterFileStorage import *



"""
Author: Matthew Turk <matthewturk@gmail.com>
Affiliation: KIPAC/SLAC/Stanford
Homepage: http://yt.enzotools.org/
License:
  Copyright (C) 2010 Matthew Turk.  All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import numpy as na
import types, os
from yt.raven import FixedResolutionBuffer, ObliqueFixedResolutionBuffer
from yt.lagos import data_object_registry, AMRProjBase, AMRSliceBase, \
                     x_dict, y_dict

class VariableMeshPanner(object):
    _buffer = None
    _hold = False

    def __init__(self, source, size, field, callback = None,
                 viewport_callback = None):
        """
        This class describes a meta-view onto a 2D data interface.  You provide
        it with a slice or a projection (*source*), a tuple of the x and y
        resolution of the final image you want (*size) and a *field* to
        display, and then it will generate a new buffer each time you call its
        methods for moving the window around and changing its size.  *callback*
        is a routine called with the new buffer each time the buffer changes
        and *viewport_callback* is called with the new *xlim* and *ylim* values
        each time the viewport changes.
        """
        #if not isinstance(source, (AMRProjBase, AMRSliceBase)):
        #    raise RuntimeError
        if callback is None:
            callback = lambda a: None
        self.callback = callback
        if viewport_callback is None:
            viewport_callback = lambda a, b: None
        self.viewport_callback = viewport_callback
        self.size = size
        self.source = source
        self.field = field
        self.xlim, self.ylim = self.bounds

    def _run_callbacks(self):
        if self._hold: return
        self.callback(self.buffer)
        self.viewport_callback(self.xlim, self.ylim)

    @property
    def bounds(self):
        if not hasattr(self, 'pf'): self.pf = self.source.pf
        DLE, DRE = self.pf["DomainLeftEdge"], self.pf["DomainRightEdge"]
        ax = self.source.axis
        xax, yax = x_dict[ax], y_dict[ax]
        xbounds = DLE[xax], DRE[xax]
        ybounds = DLE[yax], DRE[yax]
        return (xbounds, ybounds)

    @property
    def width(self):
        Wx = self.xlim[1] - self.xlim[0]
        Wy = self.ylim[1] - self.ylim[0]
        return (Wx, Wy)

    def zoom(self, factor):
        """
        This zooms the window by *factor*.
        """
        Wx, Wy = self.width
        centerx = self.xlim[0] + Wx*0.5
        centery = self.ylim[0] + Wy*0.5
        nWx, nWy = Wx/factor, Wy/factor
        self.xlim = (centerx - nWx*0.5, centerx + nWx*0.5)
        self.ylim = (centery - nWy*0.5, centery + nWy*0.5)
        self._run_callbacks()

    def pan(self, deltas):
        """
        This accepts a tuple of *deltas*, composed of (delta_x, delta_y) that
        will pan the window by those values in absolute coordinates.
        """
        self.xlim = (self.xlim[0] + deltas[0], self.xlim[1] + deltas[0])
        self.ylim = (self.ylim[0] + deltas[1], self.ylim[1] + deltas[1])
        self._run_callbacks()

    def pan_x(self, delta):
        """
        This pans the window by *delta* in the x direction.
        """
        self.pan( (delta, 0.0) )

    def pan_y(self, delta):
        """
        This pans the window by *delta* in the y direction.
        """
        self.pan( (0.0, delta) )

    def pan_rel(self, deltas):
        """
        This accepts a tuple of *deltas*, composed of (delta_x, delta_y) that
        will pan the window by those values in relative values, relative to the
        current window view.
        """
        Wx, Wy = self.width
        px = deltas[0] * Wx
        py = deltas[1] * Wy
        self.pan( (px, py) )

    def pan_rel_x(self, delta):
        """
        This pans the window by *delta* in the x direction, where *delta* is
        relative to the current window view.
        """
        self.pan_rel( (delta, 0.0) )

    def pan_rel_y(self, delta):
        """
        This pans the window by *delta* in the y direction, where *delta* is
        relative to the current window view.
        """
        self.pan_rel( (0.0, delta) )

    @property
    def buffer(self):
        my_bounds = self.xlim + self.ylim # Tuples concatenate
        # If the buffer is None, we just kill it regardless.
        if self._buffer is None:
            self._regenerate_buffer()
        elif self._buffer.bounds != my_bounds:
            self._regenerate_buffer()
        return self._buffer[self.field]

    def _regenerate_buffer(self):
        new_buffer = FixedResolutionBuffer(
            self.source, self.xlim + self.ylim,
            self.size)
        self._buffer = new_buffer

    def set_low_high(self, low, high):
        """
        This accepts *low* in the format (x_min, y_min) and *high* in the
        format (x_max, y_max) and changes the viewport appropriately.  In
        breaking from tradition, it also returns the buffer.
        """
        self.xlim = (low[0], high[0])
        self.ylim = (low[1], high[1])
        return na.log10(self.buffer)

    def set_limits(self, xlim, ylim):
        """
        This accepts a new *xlim* and *ylim*.
        """
        self.xlim = xlim
        self.ylim = ylim
        self._run_callbacks()

data_object_registry["image_panner"] = VariableMeshPanner

class WindowedVariableMeshPanner(VariableMeshPanner):

    def __init__(self, source, full_size, my_size, start_indices,
                 field, callback = None):
        """
        This image panner accepts a *full_size*, which describes the full size
        of the image which it will be displaying a portion of.  *my_size* is
        the size that this window will be responsible for, and *start_indices*
        is a tuple of indices that this window begins at.  *field* and
        *callback* function as in the vanilla image panner.
        """
        self.my_size = my_size
        self.start_indices = start_indices
        VariableMeshPanner.__init__(self, source, full_size, field, callback)

    def _regenerate_buffer(self):
        dx = (self.xlim[1] - self.xlim[0])/self.size[0]
        dy = (self.ylim[1] - self.ylim[0])/self.size[1]
        my_lim = (self.xlim[0] + dx*self.start_indices[0],
                  self.xlim[0] + dx*(self.start_indices[0] + self.my_size[0]),
                  self.ylim[0] + dx*self.start_indices[1],
                  self.ylim[0] + dx*(self.start_indices[1] + self.my_size[1]))
        new_buffer = FixedResolutionBuffer(self.source, my_lim, self.my_size)
        self._buffer = new_buffer

data_object_registry["windowed_image_panner"] = WindowedVariableMeshPanner

class MultipleWindowVariableMeshPanner(object):
    def __init__(self, windows):
        """
        This panner is exclusively a controller.  It accepts a list of
        *windows*, to which it will issue commands.  It knows nothing other
        than their interfaces.
        """
        self.windows = windows

    def zoom(self, factor):
        for w in self.windows: w.zoom(factor)

    def pan(self, deltas):
        for w in self.windows: w.pan(factor)

    def pan_x(self, delta):
        for w in self.windows: w.pan_x(delta)

    def pan_y(self, delta):
        for w in self.windows: w.pan_y(delta)

    def pan_rel(self, deltas):
        for w in self.windows: w.pan_rel(deltas)

    def pan_rel_x(self, delta):
        for w in self.windows: w.pan_rel_x(delta)

    def pan_rel_y(self, delta):
        for w in self.windows: w.pan_rel_y(delta)

    def set_limits(self, xlim, ylim):
        for w in self.windows: w.set_limits(xlim, ylim)

class RemoteWindowedVariableMeshController(MultipleWindowVariableMeshPanner):
    def __init__(self, source, mec = None):
        """
        This panner controls remote windowed panners.  It requires a *source*,
        which will be pickled and sent to the remote panners, which it will
        create as requested.  If not supplied with a *mec* (an IPython
        MultiEngineClient) it will create one itself.
        """
        if mec is None:
            from IPython.kernel.client import get_multiengine_client
            mec = get_multiengine_client()
        self.mec = mec
        self.mec.execute("import yt.extensions.image_panner")
        self._var_name = "_image_panner_%s" % (id(self))
        self._pf_name = "_pf_%s" % (id(self))
        self._source_name = "_source_%s" % (id(self))
        self.source = source
        self.mec.execute("from yt.mods import *")
        self.mec.execute("from yt.funcs import iterable")
        self.mec.push({self._pf_name: self.pf})
        self.mec.execute("%s.h" % self._pf_name)
        self.mec.push({self._source_name: self.source})
        # Now, because the double pickling tosses a PF hash reference inside
        # the unpickled object, we work around it a little
        self.mec.execute("while iterable(%s): %s = %s[1]" % (
            self._source_name, self._source_name, self._source_name))
        self.windows = []

    def add_window(self, *args, **kwargs):
        """
        This will create a new remote WindowedVariableMeshImagePanner.  The
        *args* and *kwargs* supplied here will be passed over, but the *source*
        argument is implicitly handled by this routine.
        """
        engine_id = len(self.windows)
        an = "_args_%s" % id(self)
        kn = "_kwargs_%s" % id(self)
        if 'callback' not in kwargs:
            kwargs['callback'] = ImageSaver(engine_id)
        self.mec.push({an: args, kn: kwargs}, engine_id)
        exec_string = "%s = %s.h.windowed_image_panner(%s, *%s, **%s)" % (
            self._var_name, self._pf_name, self._source_name, an, kn)
        self.mec.execute(exec_string, engine_id)
        self.windows.append(WindowedVariableMeshPannerProxy(
            self.mec, engine_id, self._var_name, id(self)))

data_object_registry["remote_image_panner"] = RemoteWindowedVariableMeshController

_wrapped_methods = ["zoom", "pan", "pan_x", "pan_y", "pan_rel",
                     "pan_rel_x", "pan_rel_y", "set_limits"]

class WindowedVariableMeshPannerProxy(object):
    class __metaclass__(type):
        def __new__(cls, name, b, d):
            # We add on a bunch of proxy functions
            def return_proxy(fname):
                def func(self, *args, **kwargs):
                    vn = "_ret_%s" % self._cid
                    an = "_args_%s" % self._cid
                    kn = "_kwargs_%s" % self._cid
                    self.mec.push({an: args, kn: kwargs}, self.engine_id)
                    exec_string = "%s = %s.%s(*%s, **%s)" % (
                        vn, self._var_name, fname, an, kn)
                    print "Executing %s on %s" % (exec_string, self.engine_id)
                    self.mec.execute(exec_string, self.engine_id)
                    return self.mec.pull(vn, self.engine_id)
                return func
            new_dict = {}
            new_dict.update(d)
            for f in _wrapped_methods:
                new_dict[f] = return_proxy(f)
            return type.__new__(cls, name, b, new_dict)

    def __init__(self, mec, engine_id, var_name, cid):
        # mec here is, optionally, an instance of MultiEngineClient
        self._var_name = var_name
        self._cid = cid
        self.engine_id = engine_id
        self.mec = mec

    @property
    def bounds(self):
        vn = "_ret_%s" % self._cid
        self.mec.execute("%s = %s.bounds" % (vn, self._var_name),
                         self.engine_id)
        return self.mec.pull(vn, self.engine_id)

    @property
    def width(self):
        vn = "_ret_%s" % self._cid
        self.mec.execute("%s = %s.width" % (vn, self._var_name),
                         self.engine_id)
        return self.mec.pull(vn, self.engine_id)

    @property
    def buffer(self):
        vn = "_ret_%s" % self._cid
        self.mec.execute("%s = %s.buffer" % (vn, self._var_name),
                         self.engine_id)
        return self.mec.pull(vn, self.engine_id)

    def _regenerate_buffer(self):
        return

    def _run_callback(self):
        self.mec.execute("%s._regenerate_buffer()" % self._var_name,
                         self.engine_id)
        self.mec.execute("%s.callback(%s.buffer)" % (
            self._var_name, self._var_name), self.engine_id)

class ProxySource(object):
    # This proxies only the things we know we need
    # Note that we assume we will only have a single engine.
    def __init__(self, mec, idnum, source_varname):
        self.mec = mec
        self.idnum = idnum
        self.source_varname = source_varname
        self.mec.execute("_tmp_%s = %s.axis" % (
            self.idnum, self.source_varname))
        self.axis = self.mec.pull("_tmp_%s" % self.idnum)[0]

    def keys(self):
        self.mec.execute("_tmp_%s = %s.keys()" % (
            self.idnum, self.source_varname))
        keys = self.mec.pull("_tmp_%s" % self.idnum)[0]
        dd = dict( (k, None) for k in keys )
        return dd

    @property
    def pf(self):
        self.mec.execute("_tmp_%s = %s.pf['DomainLeftEdge']" % (
            self.idnum, self.source_varname))
        DLE = self.mec.pull("_tmp_%s" % self.idnum)[0]
        self.mec.execute("_tmp_%s = %s.pf['DomainRightEdge']" % (
            self.idnum, self.source_varname))
        DRE = self.mec.pull("_tmp_%s" % self.idnum)[0]
        return dict(DomainLeftEdge = DLE, DomainRightEdge = DRE)

class ProxyFixedResolutionBuffer(dict):
    pass

class NonLocalDataImagePanner(VariableMeshPanner):
    def __init__(self, mec, source_varname, size, field,
                 callback = None, viewport_callback = None):
        self.source_varname = source_varname
        self._var_name = "_image_panner_%s" % (id(self))
        self.mec = mec
        self.mec.execute("import yt.extensions.image_panner")
        self.mec.execute("%s = yt.extensions.image_panner.VariableMeshPanner(" % (
                        self._var_name) +
                          "%s, (%s, %s), '%s')" % (
                        source_varname, size[0], size[1], field))

        ps = ProxySource(mec, id(self), source_varname)
        self._prfb = ProxyFixedResolutionBuffer()

        VariableMeshPanner.__init__(self, ps, size, field,
                        callback, viewport_callback)

    def _regenerate_buffer(self):
        args = (self.xlim, self.ylim)
        self.mec.push({'_tmp_%s' % id(self) : args}, block=False)
        self.mec.execute("%s.set_limits(*_tmp_%s)" % (self._var_name, id(self)),
                         block=False)
        self.mec.execute("_tmp_%s = %s.buffer" % (id(self), self._var_name),
                         block=False)
        self._prfb[self.field] = self.mec.pull("_tmp_%s" % (id(self)))[0]
        self._prfb.bounds = self.xlim + self.ylim
        self._buffer = self._prfb

class ImageSaver(object):
    def __init__(self, tile_id):
        """
        This is a sample callback for an image panner that will save to a file
        named ``wimage_%03i.png`` where the final variable is specified here as
        *tile_id*.
        """
        self.tile_id = tile_id

        import matplotlib;matplotlib.use("Agg");import pylab
        self.pylab = pylab
        self.pylab.clf()
        fig = pylab.gcf()
        fig.subplots_adjust(left=0.0, bottom=0.0,
                            right=1.0, top=1.0,
                            wspace=0.0, hspace=0.0)
        fig.set_dpi(100.0)
        fig.set_size_inches((5.12, 5.12))

    def __call__(self, val):
        self.pylab.clf()
        self.pylab.imshow(na.log10(val), interpolation='nearest')
        self.pylab.savefig("wimage_%03i.png" % self.tile_id)

class PanningCeleritasStreamer(object):
    _initialized = False
    def __init__(self, tile_id, cmap = "algae", port = 9988,
                 zlim = (0.0, 1.0), take_log = True):
        """
        This is an in-development mechanism for supplying buffers to a
        Celeritas server.
        """
        self.tile_id = tile_id
        self._port = port
        self.cmap = cmap
        self.zlim = zlim
        self.take_log = True

    def initialize(self, shape):
        if isinstance(self.cmap, types.StringTypes):
            import matplotlib.cm
            self.cmap = matplotlib.cm.get_cmap(self.cmap)

        import celeritas_streamer
        self.cs = celeritas_streamer.CeleritasStream()
        #print "Setting shape: %s and port: %s in %s" % (
        #    shape, self._port, os.getpid())
        self.cs.setSize(*shape)
        self.cs.setLocalPort(self._port)
        self.cs.initialize()
        self._initialized = True

    def __call__(self, val):
        if not self._initialized: self.initialize(val.shape)
        if self.take_log:
            vv = na.log10(val)
        else:
            vv = val.copy()
        na.subtract(vv, self.zlim[0], vv)
        na.divide(vv, (self.zlim[1]-self.zlim[0]), vv)
        new_buf = self.cmap(vv)[:,:,:3]
        na.multiply(new_buf, 255.0, new_buf)
        new_buf = new_buf.astype('uint8')
        self.cs.readFromRGBMemAndSend(new_buf)



"""
Some convenience functions, objects, and iterators

Author: Matthew Turk <matthewturk@gmail.com>
Affiliation: KIPAC/SLAC/Stanford
Homepage: http://yt.enzotools.org/
License:
  Copyright (C) 2007-2009 Matthew Turk.  All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import glob

# Named imports
import yt.lagos as lagos
import yt.raven as raven
from yt.funcs import *
import numpy as na
import os.path, inspect, types
from functools import wraps
from yt.logger import ytLogger as mylog
from yt.fido import output_type_registry

def all_pfs(max_depth=1, name_spec="*.hierarchy", **kwargs):
    """
    This function searchs a directory and its sub-directories, up to a depth of
    *max_depth*, for parameter files.  It looks for the *name_spec* and then
    instantiates an EnzoStaticOutput from each.  All subsequent *kwargs* are
    passed on to the EnzoStaticOutput constructor.
    """
    list_of_names = []
    for i in range(max_depth):
        bb = list('*' * i) + [name_spec]
        list_of_names += glob.glob(os.path.join(*bb))
    list_of_names.sort(key=lambda b: os.path.basename(b))
    for fn in list_of_names:
        yield lagos.EnzoStaticOutput(fn[:-10], **kwargs)

def max_spheres(width, unit, **kwargs):
    """
    This calls :func:`~yt.convenience.all_pfs` and then for each parameter file
    creates a :class:`~yt.lagos.AMRSphereBase` for each one,
    centered on the point of highest density, with radius *width* in units of
    *unit*.
    """
    for pf in all_pfs(**kwargs):
        v, c = pf.h.find_max("Density")
        yield pf.h.sphere(c, width/pf[unit])

def load(*args ,**kwargs):
    """
    This function attempts to determine the base data type of a filename or
    other set of arguments by calling
    :meth:`yt.lagos.StaticOutput._is_valid` until it finds a
    match, at which point it returns an instance of the appropriate
    :class:`yt.lagos.StaticOutput` subclass.
    """
    candidates = []
    for n, c in output_type_registry.items():
        if n is None: continue
        if c._is_valid(*args, **kwargs): candidates.append(n)
    if len(candidates) == 1:
        return output_type_registry[candidates[0]](*args, **kwargs)
    if len(candidates) == 0:
        mylog.error("Couldn't figure out output type for %s", args[0])
        return None
    mylog.error("Multiple output type candidates for %s:", args[0])
    for c in candidates:
        mylog.error("    Possible: %s", c)
    return None

def projload(pf, axis, weight_field = None):
    # This is something of a hack, so that we can just get back a projection
    # and not utilize any of the intermediate hierarchy objects.
    class ProjMock(dict):
        pass
    import h5py
    f = h5py.File(os.path.join(pf.fullpath, pf.parameter_filename + ".yt"))
    b = f["/Projections/%s/" % (axis)]
    wf = "weight_field_%s" % weight_field
    if wf not in b: raise KeyError(wf)
    fields = []
    for k in b:
        if k.startswith("weight_field"): continue
        if k.endswith("_%s" % weight_field):
            fields.append(k)
    proj = ProjMock()
    for f in ["px","py","pdx","pdy"]:
        proj[f] = b[f][:]
    for f in fields:
        new_name = f[:-(len(weight_field) + 1)]
        proj[new_name] = b[f][:]
    proj.axis = axis
    proj.pf = pf
    return proj



"""
AMR hierarchy container class

Author: Matthew Turk <matthewturk@gmail.com>
Affiliation: KIPAC/SLAC/Stanford
Homepage: http://yt.enzotools.org/
License:
  Copyright (C) 2007-2009 Matthew Turk.  All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from yt.arraytypes import *
from yt.logger import lagosLogger as mylog
from EnzoDefs import NUMTOCHECK

class ObjectFindingMixin(object):

    def find_ray_grids(self, coord, axis):
        """
        Returns the (objects, indices) of grids that an (x,y) ray intersects
        along *axis*
        """
        # Let's figure out which grids are on the slice
        mask=na.ones(self.num_grids)
        # So if gRE > coord, we get a mask, if not, we get a zero
        #    if gLE > coord, we get a zero, if not, mask
        # Thus, if the coordinate is between the two edges, we win!
        na.choose(na.greater(self.grid_right_edge[:,x_dict[axis]],coord[0]),(0,mask),mask)
        na.choose(na.greater(self.grid_left_edge[:,x_dict[axis]],coord[0]),(mask,0),mask)
        na.choose(na.greater(self.grid_right_edge[:,y_dict[axis]],coord[1]),(0,mask),mask)
        na.choose(na.greater(self.grid_left_edge[:,y_dict[axis]],coord[1]),(mask,0),mask)
        ind = na.where(mask == 1)
        return self.grids[ind], ind

    def find_max(self, field, finest_levels = True):
        """
        Returns (value, center) of location of maximum for a given field.
        """
        if (field, finest_levels) in self._max_locations:
            return self._max_locations[(field, finest_levels)]
        mg, mc, mv, pos = self.find_max_cell_location(field, finest_levels)
        self._max_locations[(field, finest_levels)] = (mv, pos)
        return mv, pos

    def find_max_cell_location(self, field, finest_levels = True):
        if finest_levels is True:
            gi = (self.grid_levels >= self.max_level - NUMTOCHECK).ravel()
            source = self.grid_collection([0.0]*3, self.grids[gi])
        else:
            source = self.all_data()
        mylog.debug("Searching %s grids for maximum value of %s",
                    len(source._grids), field)
        max_val, maxi, mx, my, mz, mg = \
            source.quantities["MaxLocation"]( field, lazy_reader=True)
        max_grid = self.grids[mg]
        mc = na.unravel_index(maxi, max_grid.ActiveDimensions)
        mylog.info("Max Value is %0.5e at %0.16f %0.16f %0.16f in grid %s at level %s %s", \
              max_val, mx, my, mz, max_grid, max_grid.Level, mc)
        self.parameters["Max%sValue" % (field)] = max_val
        self.parameters["Max%sPos" % (field)] = "%s" % ((mx,my,mz),)
        return max_grid, mc, max_val, na.array((mx,my,mz), dtype='float64')

    def find_min(self, field):
        """
        Returns (value, center) of location of minimum for a given field
        """
        gI = na.where(self.grid_levels >= 0) # Slow but pedantic
        minVal = 1e100
        for grid in self.grids[gI[0]]:
            mylog.debug("Checking %s (level %s)", grid.id, grid.Level)
            val, coord = grid.find_min(field)
            if val < minVal:
                minCoord = coord
                minVal = val
                minGrid = grid
        mc = na.array(minCoord)
        pos=minGrid.get_position(mc)
        mylog.info("Min Value is %0.5e at %0.16f %0.16f %0.16f in grid %s at level %s", \
              minVal, pos[0], pos[1], pos[2], minGrid, minGrid.Level)
        self.center = pos
        self.parameters["Min%sValue" % (field)] = minVal
        self.parameters["Min%sPos" % (field)] = "%s" % (pos)
        return minVal, pos

    def find_point(self, coord):
        """
        Returns the (objects, indices) of grids containing an (x,y,z) point
        """
        mask=na.ones(self.num_grids)
        for i in xrange(len(coord)):
            na.choose(na.greater(self.grid_left_edge[:,i],coord[i]), (mask,0), mask)
            na.choose(na.greater(self.grid_right_edge[:,i],coord[i]), (0,mask), mask)
        ind = na.where(mask == 1)
        return self.grids[ind], ind

    def find_slice_grids(self, coord, axis):
        """
        Returns the (objects, indices) of grids that a slice intersects along
        *axis*
        """
        # Let's figure out which grids are on the slice
        mask=na.ones(self.num_grids)
        # So if gRE > coord, we get a mask, if not, we get a zero
        #    if gLE > coord, we get a zero, if not, mask
        # Thus, if the coordinate is between the edges, we win!
        #ind = na.where( na.logical_and(self.grid_right_edge[:,axis] > coord, \
                                       #self.grid_left_edge[:,axis] < coord))
        na.choose(na.greater(self.grid_right_edge[:,axis],coord),(0,mask),mask)
        na.choose(na.greater(self.grid_left_edge[:,axis],coord),(mask,0),mask)
        ind = na.where(mask == 1)
        return self.grids[ind], ind

    def find_sphere_grids(self, center, radius):
        """
        Returns objects, indices of grids within a sphere
        """
        centers = (self.grid_right_edge + self.grid_left_edge)/2.0
        long_axis = na.maximum.reduce(self.grid_right_edge - self.grid_left_edge, 1)
        t = na.abs(centers - center)
        DW = self.parameter_file["DomainRightEdge"] \
           - self.parameter_file["DomainLeftEdge"]
        na.minimum(t, na.abs(DW-t), t)
        dist = na.sqrt(na.sum((t**2.0), axis=1))
        gridI = na.where(dist < (radius + long_axis))
        return self.grids[gridI], gridI

    def get_box_grids(self, left_edge, right_edge):
        """
        Gets back all the grids between a left edge and right edge
        """
        grid_i = na.where((na.all(self.grid_right_edge > left_edge, axis=1)
                         & na.all(self.grid_left_edge < right_edge, axis=1)) == True)
        return self.grids[grid_i], grid_i

    def get_periodic_box_grids(self, left_edge, right_edge):
        left_edge = na.array(left_edge)
        right_edge = na.array(right_edge)
        mask = na.zeros(self.grids.shape, dtype='bool')
        dl = self.parameters["DomainLeftEdge"]
        dr = self.parameters["DomainRightEdge"]
        db = right_edge - left_edge
        for off_x in [-1, 0, 1]:
            nle = left_edge.copy()
            nre = left_edge.copy()
            nle[0] = dl[0] + (dr[0]-dl[0])*off_x + left_edge[0]
            for off_y in [-1, 0, 1]:
                nle[1] = dl[1] + (dr[1]-dl[1])*off_y + left_edge[1]
                for off_z in [-1, 0, 1]:
                    nle[2] = dl[2] + (dr[2]-dl[2])*off_z + left_edge[2]
                    nre = nle + db
                    g, gi = self.get_box_grids(nle, nre)
                    mask[gi] = True
        return self.grids[mask], na.where(mask)




"""
HaloProfiler class and member functions.

Author: Britton Smith <brittons@origins.colorado.edu>
Affiliation: CASA/University of CO, Boulder
Homepage: http://yt.enzotools.org/
License:
  Copyright (C) 2008-2009 Britton Smith.  All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import yt.lagos as lagos
from yt.lagos.HaloFinding import HaloFinder
from yt.logger import lagosLogger as mylog
import yt.raven as raven
from halo_filters import *
import numpy as na
import os
import h5py
import types

PROFILE_RADIUS_THRESHOLD = 2

class HaloProfiler(lagos.ParallelAnalysisInterface):
    "Radial profiling, filtering, and projections for halos in cosmological simulations."
    def __init__(self, dataset, halos='multiple', halo_list_file='HopAnalysis.out', halo_list_format='yt_hop',
                 halo_finder_function=HaloFinder, halo_finder_args=None, halo_finder_kwargs=None,
                 use_density_center=False, density_center_exponent=1.0, use_field_max_center=None,
                 halo_radius=0.1, radius_units='1', n_profile_bins=50,
                 profile_output_dir='radial_profiles', projection_output_dir='projections',
                 projection_width=8.0, projection_width_units='mpc', project_at_level='max',
                 velocity_center=['bulk', 'halo'], filter_quantities=['id','center']):
        """
        Initialize a HaloProfiler object.
        :param halos (str): "multiple" for profiling more than one halo.  In this mode halos are read in 
               from a list or identified with a halo finder.  In "single" mode, the one and only halo 
               center is identified automatically as the location of the peak in the density field.  
               Default: "multiple".
        :param halo_list_file (str): name of file containing the list of halos.  The HaloProfiler will 
               look for this file in the data directory.  Default: "HopAnalysis.out".
        :param halo_list_format (str or dict): the format of the halo list file.  "yt_hop" for the format 
               given by yt's halo finders.  "enzo_hop" for the format written by enzo_hop.  This keyword 
               can also be given in the form of a dictionary specifying the column in which various 
               properties can be found.  For example, {"id": 0, "center": [1, 2, 3], "mass": 4, "radius": 5}.  
               Default: "yt_hop".
        :param halo_finder_function (function): If halos is set to multiple and the file given by 
               halo_list_file does not exit, the halo finding function specified here will be called.  
               Default: HaloFinder (yt_hop).
        :param halo_finder_args (tuple): args given with call to halo finder function.  Default: None.
        :param halo_finder_kwargs (dict): kwargs given with call to halo finder function. Default: None.
        :param use_density_center (bool): re-center halos before performing profiles with an center of mass 
               weighted by overdensity.  This is generally not needed.  Default: False.
        :param density_center_exponent (float): when use_density_center set to True, this specifies the 
               exponent, alpha, such that the halo center calculation is weighted by overdensity^alpha.  
               Default: 1.0.
        :param use_field_max_center (str): another alternative for halo re-centering by selecting the 
               location of the maximum of the field given by this keyword.  This is generally not needed.  
               Default: None.
        :param halo_radius (float): if no halo radii are provided in the halo list file, this parameter is 
               used to specify the radius out to which radial profiles will be made.  This keyword is also 
               used when halos is set to single.  Default: 0.1.
        :param radius_units (str): the units of halo_radius.  Default: "1" (code units).
        :param n_profile_bins (int): the number of bins in the radial profiles.  Default: 50.
        :param profile_output_dir (str): the subdirectory, inside the data directory, in which radial profile 
               output files will be created.  The directory will be created if it does not exist.  
               Default: "radial_profiles".
        :param projection_output_dir (str): the subdirectory, inside the data directory, in which projection 
               output files will be created.  The directory will be created if it does not exist.  
               Default: "projections".
        :param projection_width (float): the width of halo projections.  Default: 8.0.
        :param projection_width_units (str): the units of projection_width. Default: "mpc".
        :param project_at_level (int or "max"): the maximum refinement level to be included in projections.  
               Default: "max" (maximum level within the dataset).
        :param velocity_center (list): the method in which the halo bulk velocity is calculated (used for 
               calculation of radial and tangential velocities.  Valid options are:
     	          - ["bulk", "halo"] (Default): the velocity provided in the halo list
                  - ["bulk", "sphere"]: the bulk velocity of the sphere centered on the halo center.
    	          - ["max", field]: the velocity of the cell that is the location of the maximum of the field 
                                    specified (used only when halos set to single).
        :param filter_quantities (list): quantities from the original halo list file to be written out in the 
               filtered list file.  Default: ['id','center'].
        """

        self.dataset = dataset

        self.profile_output_dir = profile_output_dir
        self.projection_output_dir = projection_output_dir
        self.n_profile_bins = n_profile_bins
        self.projection_width = projection_width
        self.projection_width_units = projection_width_units
        self.project_at_level = project_at_level
        self.filter_quantities = filter_quantities
        if self.filter_quantities is None: self.filter_quantities = []

        self.profile_fields = []
        self.projection_fields = []

        self._halo_filters = []
        self.all_halos = []
        self.filtered_halos = []
        self._projection_halo_list = []

        # Set halo finder function and parameters, if needed.
        self.halo_finder_function = halo_finder_function
        self.halo_finder_args = halo_finder_args
        if self.halo_finder_args is None: self.halo_finder_args = ()
        self.halo_finder_kwargs = halo_finder_kwargs
        if self.halo_finder_kwargs is None: self.halo_finder_kwargs = {}

        # Set option to get halos from hop or single halo at density maximum.
        # multiple: get halos from hop
        # single: get single halo from density maximum
        self.halos = halos
        if not(self.halos is 'multiple' or self.halos is 'single'):
            mylog.error("Keyword, halos, must be either 'single' or 'multiple'.")
            return None

        # Set halo list format.
        # 'yt_hop': yt hop output.
        # 'enzo_hop': enzo_hop output.
        # dictionary: a dictionary containing fields and their corresponding columns.
        self.halo_list_file = halo_list_file
        if halo_list_format == 'yt_hop':
            self.halo_list_format = {'id':0, 'mass':1, 'center':[7, 8, 9], 'velocity':[10, 11, 12], 'r_max':13}
        elif halo_list_format == 'enzo_hop':
            self.halo_list_format = {'id':0, 'center':[4, 5, 6]}
        elif isinstance(halo_list_format, types.DictType):
            self.halo_list_format = halo_list_format
        else:
            mylog.error("Keyword, halo_list_format, must be 'yt_hop', 'enzo_hop', or a dictionary of custom settings.")
            return None

        # Option to recenter sphere on density center.
        self.use_density_center = use_density_center
        self.density_center_exponent = density_center_exponent
        if self.use_density_center:
            def _MatterDensityXTotalMass(field, data):
                return na.power((data['Matter_Density'] * data['TotalMassMsun']), self.density_center_exponent)
            def _Convert_MatterDensityXTotalMass(data):
                return 1
            lagos.add_field("MatterDensityXTotalMass", units=r"",
                      function=_MatterDensityXTotalMass,
                      convert_function=_Convert_MatterDensityXTotalMass)

        # Option to recenter sphere on the location of a field max.
        self.use_field_max_center = use_field_max_center
        if self.use_field_max_center is not None:
            self.use_density_center = False

        # Look for any field that might need to have the bulk velocity set.
        self._need_bulk_velocity = False
        for field in [hp['field'] for hp in self.profile_fields]:
            if 'Velocity' in field or 'Mach' in field:
                self._need_bulk_velocity = True
                break

        # Check validity for VelocityCenter parameter which toggles how the 
        # velocity is zeroed out for radial velocity profiles.
        self.velocity_center = velocity_center[:]
        if self.velocity_center[0] == 'bulk':
            if self.velocity_center[1] == 'halo' and \
                    self.halos is 'single':
                mylog.error("Parameter, VelocityCenter, must be set to 'bulk sphere' or 'max <field>' with halos flag set to 'single'.")
                return None
            if self.velocity_center[1] == 'halo' and \
                    self.halo_list_format is 'enzo_hop':
                mylog.error("Parameter, VelocityCenter, must be 'bulk sphere' for old style hop output files.")
                return None
            if not(self.velocity_center[1] == 'halo' or 
                   self.velocity_center[1] == 'sphere'):
                mylog.error("Second value of VelocityCenter must be either 'halo' or 'sphere' if first value is 'bulk'.")
                return None
        elif self.velocity_center[0] == 'max':
            if self.halos is 'multiple':
                mylog.error("Getting velocity center from a max field value only works with halos='single'.")
                return None
        else:
            mylog.error("First value of parameter, VelocityCenter, must be either 'bulk' or 'max'.")
            return None

        # Create dataset object.
        self.pf = lagos.EnzoStaticOutput(self.dataset)
        self.pf.h
        if self.halos is 'single' or not 'r_max' in self.halo_list_format:
            self.halo_radius = halo_radius / self.pf[radius_units]

        # Get halo(s).
        if self.halos is 'single':
            v, center = self.pf.h.find_max('Density')
            singleHalo = {}
            singleHalo['center'] = center
            singleHalo['r_max'] = self.halo_radius * self.pf.units['mpc']
            singleHalo['id'] = 0
            self.all_halos.append(singleHalo)
        elif self.halos is 'multiple':
            # Get hop data.
            self._load_halo_data()
            if len(self.all_halos) == 0:
                mylog.error("No halos loaded, there will be nothing to do.")
                return None
        else:
            mylog.error("I don't know whether to get halos from hop or from density maximum.  This should not have happened.")
            return None

    def add_halo_filter(self, function, *args, **kwargs):
        "Add a halo filter to the filter list."

        self._halo_filters.append({'function':function, 'args':args, 'kwargs':kwargs})

    def add_profile(self, field, weight_field=None, accumulation=False):
        "Add a field for profiling."

        self.profile_fields.append({'field':field, 'weight_field':weight_field, 'accumulation':accumulation})

    def add_projection(self, field, weight_field=None):
        "Add a field for projection."

        self.projection_fields.append({'field':field, 'weight_field':weight_field})

    @lagos.parallel_blocking_call
    def make_profiles(self, filename=None, prefilters=None, **kwargs):
        "Make radial profiles for all halos on the list."

        # Reset filtered halo list.
        self.filtered_halos = []

        # Check to see if the VirialFilter has been added to the filter list.
        # If a lower mass cutoff is being used, use it to make a pre-filter.
        if prefilters is None: prefilters = []
        virial_prefilter = None
        virial_prefilter_safety_factor = 0.5
        all_filter_functions = [hf['function'] for hf in self._halo_filters]
        virial_filter = VirialFilter in all_filter_functions
        if 'mass' in self.halo_list_format and VirialFilter in all_filter_functions:
            vFilter = self._halo_filters[all_filter_functions.index(VirialFilter)]
            if vFilter['kwargs'].has_key('virial_filters') and \
               vFilter['kwargs']['virial_filters'] is not None:
                all_vqFilters = [vqf[0] for vqf in vFilter['kwargs']['virial_filters']]
                if 'TotalMassMsun' in all_vqFilters:
                    mass_filter = vFilter['kwargs']['virial_filters'][all_vqFilters.index('TotalMassMsun')]
                    if '>' in mass_filter[1]:
                        virial_prefilter = "halo['mass'] %s %f * %s" % (mass_filter[1], virial_prefilter_safety_factor, mass_filter[2])
                        prefilters.append(virial_prefilter)
                    elif '<' in mass_filter[1]:
                        virial_prefilter = "halo['mass'] %s %f * %s" % (mass_filter[1], (1./virial_prefilter_safety_factor), mass_filter[2])
                        prefilters.append(virial_prefilter)

        # Add profile fields necessary for calculating virial quantities.
        if virial_filter: self._check_for_needed_profile_fields()

        outputDir = "%s/%s" % (self.pf.fullpath, self.profile_output_dir)
        self.__check_directory(outputDir)

        # Profile all halos.
        for halo in self._get_objs('all_halos', round_robin=True):

            # Apply prefilters to avoid profiling unwanted halos.
            filter_result = True
            haloQuantities = {}
            if prefilters is not None:
                for prefilter in prefilters:
                    if not eval(prefilter):
                        filter_result = False
                        break

            if filter_result and len(self.profile_fields) > 0:

                profile_filename = "%s/Halo_%04d_profile.dat" % (outputDir, halo['id'])

                profiledHalo = self._get_halo_profile(halo, profile_filename, virial_filter=virial_filter)

                if profiledHalo is None:
                    continue

                # Apply filter and keep track of the quantities that are returned.
                for hFilter in self._halo_filters:
                    filter_result, filterQuantities = hFilter['function'](profiledHalo, *hFilter['args'], **hFilter['kwargs'])

                    if not filter_result: break

                    if filterQuantities is not None:
                        haloQuantities.update(filterQuantities)

            if filter_result:
                for quantity in self.filter_quantities:
                    if halo.has_key(quantity): haloQuantities[quantity] = halo[quantity]

                self.filtered_halos.append(haloQuantities)

        self.filtered_halos = self._mpi_catlist(self.filtered_halos)
        self.filtered_halos.sort(key = lambda a:a['id'])

        if filename is not None:
            self._write_filtered_halo_list(filename, **kwargs)

    def _get_halo_profile(self, halo, filename, virial_filter=True, force_write=False):
        """
        Profile a single halo and write profile data to a file.
        If file already exists, read profile data from file.
        Return a dictionary of id, center, and virial quantities if virial_filter is True.
        """

        # Read profile from file if it already exists.
        # If not, profile will be None.
        profile = self._read_profile(filename)

        # Make profile if necessary.
        newProfile = profile is None
        if newProfile:

            r_min = 2 * self.pf.h.get_smallest_dx() * self.pf['mpc']
            if (halo['r_max'] / r_min < PROFILE_RADIUS_THRESHOLD):
                mylog.error("Skipping halo with r_max / r_min = %f." % (halo['r_max']/r_min))
                return None

            sphere = self.pf.h.sphere(halo['center'], halo['r_max']/self.pf.units['mpc'])
            if len(sphere._grids) == 0: return None
            new_sphere = False

            if self.use_density_center:
                dc_x = sphere.quantities['WeightedAverageQuantity']('x', 'MatterDensityXTotalMass')
                dc_y = sphere.quantities['WeightedAverageQuantity']('y', 'MatterDensityXTotalMass')
                dc_z = sphere.quantities['WeightedAverageQuantity']('z', 'MatterDensityXTotalMass')
                mylog.info("Moving halo center from %s to %s." % (halo['center'], [dc_x, dc_y, dc_z]))
                halo['center'] = [dc_x, dc_y, dc_z]
                new_sphere = True

            if self.use_field_max_center is not None:
                ma, maxi, mx, my, mz, mg = sphere.quantities['MaxLocation'](self.use_field_max_center)
                mylog.info("Moving halo center from %s to %s." % (halo['center'], [mx, my, mz]))
                halo['center'] = [mx, my, mz]
                new_sphere = True

            if new_sphere:
                # Temporary solution to memory leak.
                for g in self.pf.h.grids:
                    g.clear_data()
                sphere.clear_data()
                del sphere
                sphere = self.pf.h.sphere(halo['center'], halo['r_max']/self.pf.units['mpc'])

            if self._need_bulk_velocity:
                # Set bulk velocity to zero out radial velocity profiles.
                if self.velocity_center[0] == 'bulk':
                    if self.velocity_center[1] == 'halo':
                        sphere.set_field_parameter('bulk_velocity', halo['velocity'])
                    elif self.velocity_center[1] == 'sphere':
                        sphere.set_field_parameter('bulk_velocity', sphere.quantities['BulkVelocity']())
                    else:
                        mylog.error("Invalid parameter: VelocityCenter.")
                elif self.velocity_center[0] == 'max':
                    max_grid, max_cell, max_value, max_location = self.pf.h.find_max_cell_location(self.velocity_center[1])
                    sphere.set_field_parameter('bulk_velocity', [max_grid['x-velocity'][max_cell],
                                                                 max_grid['y-velocity'][max_cell],
                                                                 max_grid['z-velocity'][max_cell]])

            profile = lagos.BinnedProfile1D(sphere, self.n_profile_bins, "RadiusMpc",
                                            r_min, halo['r_max'],
                                            log_space=True, lazy_reader=False)
            for hp in self.profile_fields:
                profile.add_fields(hp['field'], weight=hp['weight_field'], accumulation=hp['accumulation'])

        if virial_filter:
            self._add_actual_overdensity(profile)

        if newProfile:
            mylog.info("Writing halo %d" % halo['id'])
            profile.write_out(filename, format='%0.6e')
        elif force_write:
            mylog.info("Re-writing halo %d" % halo['id'])
            self._write_profile(profile, filename, format='%0.6e')

        if newProfile:
            # Temporary solution to memory leak.
            for g in self.pf.h.grids:
                g.clear_data()
            sphere.clear_data()
            del sphere

        return profile

    @lagos.parallel_blocking_call
    def make_projections(self, axes=[0, 1, 2], halo_list='filtered', save_images=False, save_cube=True):
        "Make projections of all halos using specified fields."

        # Get list of halos for projecting.
        if halo_list == 'filtered':
            self._halo_projection_list = self.filtered_halos
        elif halo_list == 'all':
            self._halo_projection_list = self.all_halos
        elif isinstance(halo_list, types.StringType):
            self._halo_projection_list = self._read_halo_list(halo_list)
        elif isinstance(halo_list, types.ListType):
            self._halo_projection_list = halo_list
        else:
            mylog.error("Keyword, halo_list', must be 'filtered', 'all', a filename, or an actual list.")
            return

        if len(self._halo_projection_list) == 0:
            mylog.error("Halo list for projections is empty.")
            return

        # Set resolution for fixed resolution output.
        if save_cube:
            if self.project_at_level == 'max':
                proj_level = self.pf.h.max_level
            else:
                proj_level = int(self.project_at_level)
            proj_dx = self.pf.units[self.projection_width_units] / self.pf.parameters['TopGridDimensions'][0] / \
                (self.pf.parameters['RefineBy']**proj_level)
            projectionResolution = int(self.projection_width / proj_dx)

        outputDir = "%s/%s" % (self.pf.fullpath, self.projection_output_dir)
        self.__check_directory(outputDir)

        center = [0.5 * (self.pf.parameters['DomainLeftEdge'][w] + self.pf.parameters['DomainRightEdge'][w])
                  for w in range(self.pf.parameters['TopGridRank'])]

        for halo in self._get_objs('_halo_projection_list', round_robin=True):
            if halo is None:
                continue
            # Check if region will overlap domain edge.
            # Using non-periodic regions is faster than using periodic ones.
            leftEdge = [(halo['center'][w] - 0.5 * self.projection_width/self.pf.units[self.projection_width_units])
                        for w in range(len(halo['center']))]
            rightEdge = [(halo['center'][w] + 0.5 * self.projection_width/self.pf.units[self.projection_width_units])
                         for w in range(len(halo['center']))]

            mylog.info("Projecting halo %04d in region: [%f, %f, %f] to [%f, %f, %f]." %
                       (halo['id'], leftEdge[0], leftEdge[1], leftEdge[2], rightEdge[0], rightEdge[1], rightEdge[2]))

            need_per = False
            for w in range(len(halo['center'])):
                if ((leftEdge[w] < self.pf.parameters['DomainLeftEdge'][w]) or
                    (rightEdge[w] > self.pf.parameters['DomainRightEdge'][w])):
                    need_per = True
                    break

            if need_per:
                region = self.pf.h.periodic_region(halo['center'], leftEdge, rightEdge)
            else:
                region = self.pf.h.region(halo['center'], leftEdge, rightEdge)

            # Make projections.
            if not isinstance(axes, types.ListType): axes = list([axes])
            for w in axes:
                # Create a plot collection.
                pc = raven.PlotCollection(self.pf, center=center)
                # YT projections do not follow the right-hand rule.
                coords = range(3)
                del coords[w]
                x_axis = coords[0]
                y_axis = coords[1]

                for hp in self.projection_fields:
                    pc.add_projection(hp['field'], w, weight_field=hp['weight_field'], data_source=region)
                
                # Set x and y limits, shift image if it overlaps domain boundary.
                if need_per:
                    pw = self.projection_width/self.pf.units[self.projection_width_units]
                    shift_projections(self.pf, pc, halo['center'], center, w)
                    # Projection has now been shifted to center of box.
                    proj_left = [center[x_axis]-0.5*pw, center[y_axis]-0.5*pw]
                    proj_right = [center[x_axis]+0.5*pw, center[y_axis]+0.5*pw]
                else:
                    proj_left = [leftEdge[x_axis], leftEdge[y_axis]]
                    proj_right = [rightEdge[x_axis], rightEdge[y_axis]]

                pc.set_xlim(proj_left[0], proj_right[0])
                pc.set_ylim(proj_left[1], proj_right[1])

                # Save projection data to hdf5 file.
                if save_cube:
                    axis_labels = ['x', 'y', 'z']
                    dataFilename = "%s/Halo_%04d_%s_data.h5" % \
                            (outputDir, halo['id'], axis_labels[w])
                    mylog.info("Saving projection data to %s." % dataFilename)

                    output = h5py.File(dataFilename, "a")
                    # Create fixed resolution buffer for each projection and write them out.
                    for e, hp in enumerate(self.projection_fields):
                        frb = raven.FixedResolutionBuffer(pc.plots[e].data, (proj_left[0], proj_right[0], proj_left[1], proj_right[1]),
                                                          (projectionResolution, projectionResolution),
                                                          antialias=False)
                        dataset_name = "%s_%s" % (hp['field'], hp['weight_field'])
                        if dataset_name in output.listnames(): del output[dataset_name]
                        output.create_dataset(dataset_name, data=frb[hp['field']])
                    output.close()

                if save_images:
                    pc.save("%s/Halo_%04d" % (outputDir, halo['id']), force_save=True)

            del region

    def _add_actual_overdensity(self, profile):
        "Calculate overdensity from TotalMassMsun and CellVolume fields."

        if 'ActualOverdensity' in profile.keys():
            return

        rho_crit_now = 1.8788e-29 * self.pf['CosmologyHubbleConstantNow']**2.0 * \
            self.pf['CosmologyOmegaMatterNow'] # g cm^-3
        Msun2g = 1.989e33
        rho_crit = rho_crit_now * ((1.0 + self.pf['CosmologyCurrentRedshift'])**3.0)

        profile['ActualOverdensity'] = (Msun2g * profile['TotalMassMsun']) / \
            profile['CellVolume'] / rho_crit

    def _check_for_needed_profile_fields(self):
        "Make sure CellVolume and TotalMass fields are added so virial quantities can be calculated."
        all_profile_fields = [hp['field'] for hp in self.profile_fields]
        if not 'CellVolume' in all_profile_fields:
            mylog.info("Adding CellVolume field to so virial quantities can be calculated")
            self.add_profile('CellVolume', weight_field=None, accumulation=True)
        if not 'TotalMassMsun' in all_profile_fields:
            mylog.info("Adding TotalMassMsun field to so virial quantities can be calculated")
            self.add_profile('TotalMassMsun', weight_field=None, accumulation=True)

    def _load_halo_data(self, filename=None):
        "Read hop output file or run hop if it doesn't exist."

        # Don't run if hop data already loaded.
        if self.all_halos:
            return

        if filename is None:
            filename = self.halo_list_file

        hopFile = "%s/%s" % (self.pf.fullpath, filename)

        if not(os.path.exists(hopFile)):
            mylog.info("Hop file not found, running hop to get halos.")
            self._run_hop(hopFile)

        self.all_halos = self._read_halo_list(hopFile)

    def _read_halo_list(self, listFile):
        """
        Read halo list from aue file.
        Allow for columnar data in varying formats.
        """

        def __isE(arg):
            parts = arg.lower().split('e')
            if len(parts) != 2: return False
            return not (True in [q.isalpha() for q in ''.join(parts)])

        def __get_num(arg):
            if __isE(arg):
                return float(arg)
            if arg != arg.swapcase():
                return arg
            return float(arg)

        mylog.info("Reading halo information from %s." % listFile)
        haloList = []
        listLines = file(listFile)

        fields = self.halo_list_format.keys()
        getID = not 'id' in fields
        getR_max = not 'r_max' in fields

        for line in listLines:
            line = line.strip()
            if not(line.startswith('#')):
                halo = {}
                onLine = line.split()
                for field in fields:
                    if isinstance(self.halo_list_format[field], types.ListType):
                        halo[field] = [__get_num(onLine[q]) for q in self.halo_list_format[field]]
                    else:
                        halo[field] = __get_num(onLine[self.halo_list_format[field]])
                if getID: halo['id'] = len(haloList)
                if getR_max: 
                    halo['r_max'] = self.halo_radius * self.pf.units['mpc']
                else:
                    halo['r_max'] *= self.pf.units['mpc']
                haloList.append(halo)

        mylog.info("Loaded %d halos." % (len(haloList)))
        return haloList

    def _read_profile(self, profileFile):
        "Read radial profile from file.  Return None if it doesn't have all the fields requested."

        # Check to see if file exists.
        if not os.path.exists(profileFile):
            return None

        f = open(profileFile, 'r')
        lines = f.readlines()
        f.close()

        # Get fields from header.
        header = lines.pop(0)
        header = header.strip()
        fields = header.split()
        # First string is '#'.
        fields.pop(0)

        profile = {}
        profile_obj = FakeProfile(self.pf)
        for field in fields:
            profile[field] = []

        # Check if all fields needed are present.
        all_profile_fields = [hp['field'] for hp in self.profile_fields]
        for field in all_profile_fields:
            if not field in profile:
                return None

        # Fill profile fields, skip bad values.
        for line in lines:
            line = line.strip()
            onLine = line.split()
            lineOK = True
            for value in onLine:
                if value.isalpha():
                    lineOK = False
                    break
            if lineOK:
                for q, field in enumerate(fields):
                    profile[field].append(float(onLine[q]))

        for field in fields:
            profile[field] = na.array(profile[field])

        profile_obj._data = profile

        if len(profile[fields[0]]) > 1:
            return profile_obj
        else:
            return None

    @lagos.parallel_blocking_call
    def _run_hop(self, hopFile):
        "Run hop to get halos."

        hop_results = self.halo_finder_function(self.pf, *self.halo_finder_args, **self.halo_finder_kwargs)
        hop_results.write_out(hopFile)

        del hop_results
        self.pf.h.clear_all_data()

    @lagos.parallel_root_only
    def _write_filtered_halo_list(self, filename, format="%s"):
        """
        Write out list of filtered halos along with any quantities 
        picked up during the filtering process.
        """

        if len(self.filtered_halos) == 0:
            mylog.error("No halos in filtered list.")
            return

        filename = "%s/%s" % (self.pf.fullpath, filename)
        mylog.info("Writing filtered halo list to %s." % filename)
        file = open(filename, "w")
        fields = [field for field in sorted(self.filtered_halos[0])]
        halo_fields = []
        for halo_field in self.filter_quantities:
            if halo_field in fields:
                fields.remove(halo_field)
                halo_fields.append(halo_field)
        # Make it so number of fields in header is same as number of data columns.
        header_fields = []
        for halo_field in halo_fields:
            if isinstance(self.filtered_halos[0][halo_field], types.ListType):
                header_fields.extend(["%s[%d]" % (halo_field, q) 
                                      for q in range(len(self.filtered_halos[0][halo_field]))])
            else:
                header_fields.append(halo_field)
        file.write("# ")
        file.write("\t".join(header_fields + fields + ["\n"]))

        for halo in self.filtered_halos:
            for halo_field in halo_fields:
                if isinstance(halo[halo_field], types.ListType):
                    field_data = na.array(halo[halo_field])
                    field_data.tofile(file, sep="\t", format=format)
                else:
                    if halo_field == 'id':
                        file.write("%04d" % halo[halo_field])
                    else:
                        file.write("%s" % halo[halo_field])
                file.write("\t")
            field_data = na.array([halo[field] for field in fields])
            field_data.tofile(file, sep="\t", format=format)
            file.write("\n")
        file.close()

    def _write_profile(self, profile, filename, format="%0.16e"):
        fid = open(filename, "w")
        fields = [field for field in sorted(profile.keys()) if field != "UsedBins"]
        fid.write("\t".join(["#"] + fields + ["\n"]))
        field_data = na.array([profile[field] for field in fields])
        for line in range(field_data.shape[1]):
            field_data[:, line].tofile(fid, sep="\t", format=format)
            fid.write("\n")
        fid.close()

    @lagos.parallel_root_only
    def __check_directory(self, outputDir):
        if (os.path.exists(outputDir)):
            if not(os.path.isdir(outputDir)):
                mylog.error("Output directory exists, but is not a directory: %s." % outputDir)
                raise IOError(outputDir)
        else:
            os.mkdir(outputDir)

def shift_projections(pf, pc, oldCenter, newCenter, axis):
    """
    Shift projection data around.
    This is necessary when projecting a preiodic region.
    """
    offset = [newCenter[q]-oldCenter[q] for q in range(len(oldCenter))]
    width = [pf.parameters['DomainRightEdge'][q]-pf.parameters['DomainLeftEdge'][q] for q in range(len(oldCenter))]

    del offset[axis]
    del width[axis]

    for plot in pc.plots:
        # Get name of data field.
        other_fields = {'px':True, 'py':True, 'pdx':True, 'pdy':True, 'weight_field':True}
        for pfield in plot.data.data.keys():
            if not(other_fields.has_key(pfield)):
                field = pfield
                break

        # Shift x and y positions.
        plot['px'] += offset[0]
        plot['py'] += offset[1]

        # Wrap off-edge cells back around to other side (periodic boundary conditions).
        plot['px'][plot['px'] < 0] += width[0]
        plot['py'][plot['py'] < 0] += width[1]
        plot['px'][plot['px'] > width[0]] -= width[0]
        plot['py'][plot['py'] > width[1]] -= width[1]

        # After shifting, some cells have fractional coverage on both sides of the box.
        # Find those cells and make copies to be placed on the other side.

        # Cells hanging off the right edge.
        add_x_px = plot['px'][plot['px'] + 0.5 * plot['pdx'] > width[0]]
        add_x_px -= width[0]
        add_x_py = plot['py'][plot['px'] + 0.5 * plot['pdx'] > width[0]]
        add_x_pdx = plot['pdx'][plot['px'] + 0.5 * plot['pdx'] > width[0]]
        add_x_pdy = plot['pdy'][plot['px'] + 0.5 * plot['pdx'] > width[0]]
        add_x_field = plot[field][plot['px'] + 0.5 * plot['pdx'] > width[0]]
        add_x_weight_field = plot['weight_field'][plot['px'] + 0.5 * plot['pdx'] > width[0]]

        # Cells hanging off the left edge.
        add2_x_px = plot['px'][plot['px'] - 0.5 * plot['pdx'] < 0]
        add2_x_px += width[0]
        add2_x_py = plot['py'][plot['px'] - 0.5 * plot['pdx'] < 0]
        add2_x_pdx = plot['pdx'][plot['px'] - 0.5 * plot['pdx'] < 0]
        add2_x_pdy = plot['pdy'][plot['px'] - 0.5 * plot['pdx'] < 0]
        add2_x_field = plot[field][plot['px'] - 0.5 * plot['pdx'] < 0]
        add2_x_weight_field = plot['weight_field'][plot['px'] - 0.5 * plot['pdx'] < 0]

        # Cells hanging off the top edge.
        add_y_px = plot['px'][plot['py'] + 0.5 * plot['pdy'] > width[1]]
        add_y_py = plot['py'][plot['py'] + 0.5 * plot['pdy'] > width[1]]
        add_y_py -= width[1]
        add_y_pdx = plot['pdx'][plot['py'] + 0.5 * plot['pdy'] > width[1]]
        add_y_pdy = plot['pdy'][plot['py'] + 0.5 * plot['pdy'] > width[1]]
        add_y_field = plot[field][plot['py'] + 0.5 * plot['pdy'] > width[1]]
        add_y_weight_field = plot['weight_field'][plot['py'] + 0.5 * plot['pdy'] > width[1]]

        # Cells hanging off the bottom edge.
        add2_y_px = plot['px'][plot['py'] - 0.5 * plot['pdy'] < 0]
        add2_y_py = plot['py'][plot['py'] - 0.5 * plot['pdy'] < 0]
        add2_y_py += width[1]
        add2_y_pdx = plot['pdx'][plot['py'] - 0.5 * plot['pdy'] < 0]
        add2_y_pdy = plot['pdy'][plot['py'] - 0.5 * plot['pdy'] < 0]
        add2_y_field = plot[field][plot['py'] - 0.5 * plot['pdy'] < 0]
        add2_y_weight_field = plot['weight_field'][plot['py'] - 0.5 * plot['pdy'] < 0]

        # Add the hanging cells back to the projection data.
        plot.data['px'] = na.concatenate([plot['px'], add_x_px, add_y_px, add2_x_px, add2_y_px])
        plot.data['py'] = na.concatenate([plot['py'], add_x_py, add_y_py, add2_x_py, add2_y_py])
        plot.data['pdx'] = na.concatenate([plot['pdx'], add_x_pdx, add_y_pdx, add2_x_pdx, add2_y_pdx])
        plot.data['pdy'] = na.concatenate([plot['pdy'], add_x_pdy, add_y_pdy, add2_x_pdy, add2_y_pdy])
        plot.data[field] = na.concatenate([plot[field], add_x_field, add_y_field, add2_x_field, add2_y_field])
        plot.data['weight_field'] = na.concatenate([plot['weight_field'],
                                                    add_x_weight_field, add_y_weight_field, add2_x_weight_field, add2_y_weight_field])

        # Delete original copies of hanging cells.
        del add_x_px, add_y_px, add2_x_px, add2_y_px
        del add_x_py, add_y_py, add2_x_py, add2_y_py
        del add_x_pdx, add_y_pdx, add2_x_pdx, add2_y_pdx
        del add_x_pdy, add_y_pdy, add2_x_pdy, add2_y_pdy
        del add_x_field, add_y_field, add2_x_field, add2_y_field
        del add_x_weight_field, add_y_weight_field, add2_x_weight_field, add2_y_weight_field

class FakeProfile(lagos.ParallelAnalysisInterface):
    """
    This is used to mimic a profile object when reading profile data from disk.
    """
    def __init__(self, pf):
        self.pf = pf
        self._data = {}

    def __getitem__(self, key):
        return self._data[key]

    def keys(self):
        return self._data.keys()



"""
Basic unit test suite to run all the tests at once. Hopefully this it's
clear how to append additional tests.

You can run this using: 

$ python tests/runall.py

This should be done from the root directory of the installation.

YT can either be installed globally, or the extensions build with:

$ python setup.py build_ext --inplace
"""

import unittest

import test_lagos
import test_raven
import test_hdf5_reader

def get_suite():
    suite_l = unittest.defaultTestLoader.loadTestsFromModule(test_lagos)
    suite_r = unittest.defaultTestLoader.loadTestsFromModule(test_raven)
    suite_h = unittest.defaultTestLoader.loadTestsFromModule(test_hdf5_reader)
    suite = unittest.TestSuite([suite_l, suite_r])
    return suite

if __name__ == '__main__':
    unittest.main(defaultTest='get_suite')



"""
HOP-output data handling

Author: Matthew Turk <matthewturk@gmail.com>
Affiliation: KIPAC/SLAC/Stanford

Author: Stephen Skory <sskory@physics.ucsd.edu>
Affiliation: UC San Diego

Homepage: http://yt.enzotools.org/
License:
  Copyright (C) 2008-2009 Matthew Turk, Stephen Skory.  All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

##
## This is a temporary file, modified by Stephen Skory,
## to start work toward supporting parallel HOP studies
##

from yt.lagos.hop import *
from numpy import *

class HopList(object):
    def __init__(self, data_source, threshold=160.0,
                 dm_only = True):
        """
        Run hop on *data_source* with a given density *threshold*.  If
        *dm_only* is set, only run it on the dark matter particles, otherwise
        on all particles.  Returns an iterable collection of *HopGroup* items.
        """
        self.data_source = data_source
        self.dm_only = dm_only
        self.threshold = threshold
        self._groups = []
        self._groups2 = {}
        self._max_dens = {}
        self._max_dens2 = {}
        mylog.info("Initializing HOP")
        self.__obtain_particles()
        #self.__enlarge_data()
        #self.__slice_data()
        #self.__run_hops()
        #self.__reduce_data()
        self.__run_hop()
        mylog.info("Parsing outputs")
        #self.__parse_outputs()
        self.__parse_output()
        #self.__glue_outputs()
        mylog.debug("Finished. (%s)", len(self))

    def __obtain_particles(self):
        if self.dm_only: ii = self.__get_dm_indices()
        else: ii = slice(None)
        self.particle_fields = {}
        for field in ["particle_position_%s" % ax for ax in 'xyz'] + \
                     ["ParticleMassMsun"]:
            tot_part = self.data_source[field].size
            self.particle_fields[field] = self.data_source[field][ii]
        self._base_indices = na.arange(tot_part)[ii]

    def __enlarge_data(self):
        """
        Take the data fields and reflect them around the sides of the
        box, using periodicity.
        """
        sizetemp = self.particle_fields["particle_position_x"].size
        self.tempx = [i for i in range(3*sizetemp)]
        self.tempy = [i for i in range(3*sizetemp)]
        self.tempz = [i for i in range(3*sizetemp)]
        self.tempm = [i for i in range(3*sizetemp)]
        self.tracking = [i for i in range(3*sizetemp)]
        # reverse below 0, straight up between 0 and 1, reverse above 1
        for part,i in enumerate(self.particle_fields["particle_position_x"]):
            self.tempx[sizetemp - 1 - part] = -i
            self.tempx[sizetemp + part] = i
            self.tempx[3*sizetemp - 1 - part] = 2 - i
            self.tracking[sizetemp - 1 - part] = part
            self.tracking[sizetemp + part] = part
            self.tracking[3*sizetemp - 1 - part] = part
        for part,i in enumerate(self.particle_fields["particle_position_y"]):
            self.tempy[sizetemp - 1 - part] = -i
            self.tempy[sizetemp + part] = i
            self.tempy[3*sizetemp - 1 - part] = 2 - i
        for part,i in enumerate(self.particle_fields["particle_position_z"]):
            self.tempz[sizetemp - 1 - part] = -i
            self.tempz[sizetemp + part] = i
            self.tempz[3*sizetemp - 1 - part] = 2 - i
        for part,i in enumerate(self.particle_fields["ParticleMassMsun"]):
            self.tempm[sizetemp - 1 - part] = i
            self.tempm[sizetemp + part] = i
            self.tempm[3*sizetemp -1 - part] = i

    def __slice_data(self):
        """
        Cut up the particle data into subboxes, depending on *cuts* and the over-sampling
        parameter *padding*. This also enlarges the box such that the particles run
        from 0 to 1, which may not be neccesary.
        """
        cuts = 3
        padding = 0.2
        self.temp2x = {}
        self.temp2y = {}
        self.temp2z = {}
        self.temp2m = {}
        self.tracking2 = {}
        # loop over the sub-boxes
        for i in range(cuts):
            for j in range(cuts):
                for k in range(cuts):
                    x_t = []
                    y_t = []
                    z_t = []
                    m_t = []
                    t_t = {}
                    # define the edges
                    xleft = (1.0 * i)/cuts - padding
                    xright = (1.0 * (i+1))/cuts + padding
                    yleft = (1.0 * j)/cuts - padding
                    yright = (1.0 * (j+1))/cuts + padding
                    zleft = (1.0 * k)/cuts - padding
                    zright = (1.0 * (k+1))/cuts + padding
                    # loop over the temporary expanded fields, test to see if they're in the sub-box
                    # and then re-map to a unit sized box, we'll have inclusive boundary on the left,
                    # exclusive on the right
                    jj = 0
                    for part,ii in enumerate(self.tempx):
                        if ((ii >= xleft) and (ii < xright) and (self.tempy[part] >= yleft) and \
                        (self.tempy[part] < yright) and (self.tempz[part] >= zleft) and \
                        (self.tempz[part] < zright)):
                            x_t.append((ii-xleft)*(1.0/(xright-xleft)))
                            y_t.append((self.tempy[part]-yleft)*(1.0/(yright-yleft)))
                            z_t.append((self.tempz[part]-zleft)*(1.0/(zright-zleft)))
                            m_t.append(self.tempm[part])
                            t_t[jj] = self.tracking[part] # put real ID in t_t at jj
                            jj += 1
                    # save them to their position in the dict
                    self.temp2x[((i*cuts) + j)*cuts + k] = array(x_t)
                    self.temp2y[((i*cuts) + j)*cuts + k] = array(y_t)
                    self.temp2z[((i*cuts) + j)*cuts + k] = array(z_t)
                    self.temp2m[((i*cuts) + j)*cuts + k] = array(m_t)
                    self.tracking2[((i*cuts) + j)*cuts + k] = t_t

    def __reduce_data(self):
        """
        Take the boxes after HOP has handled them and make them small, to their true size.
        """
        cuts = 3
        padding = 0.2
        # loop over the sub-boxes
        for i in range(cuts):
            for j in range(cuts):
                for k in range(cuts):
                    xleft = (1.0 * i)/cuts - padding
                    xright = (1.0 * (i+1))/cuts + padding
                    yleft = (1.0 * j)/cuts - padding
                    yright = (1.0 * (j+1))/cuts + padding
                    zleft = (1.0 * k)/cuts - padding
                    zright = (1.0 * (k+1))/cuts + padding
                    # we're going to reduce the values of the particles to their true position
                    # with some going negative, of course
                    self.temp2x[((i*cuts) + j)*cuts + k] = self.temp2x[((i*cuts) + j)*cuts + k]*(xright-xleft)+xleft
                    self.temp2y[((i*cuts) + j)*cuts + k] = self.temp2y[((i*cuts) + j)*cuts + k]*(yright-yleft)+yleft
                    self.temp2z[((i*cuts) + j)*cuts + k] = self.temp2z[((i*cuts) + j)*cuts + k]*(zright-zleft)+zleft

    def __run_hops(self):
        """
        Run HOP on the different boxes.
        """
        cuts = 3
        padding = 0.2
        nParts = self.particle_fields["particle_position_x"].size
        dens_fix = 1. / (2*padding + 1./cuts)**3
        self.densities2 = {}
        self.tags2 = {}
        for i in range(cuts):
            for j in range(cuts):
                for k in range(cuts):
                    print "run_hops %d" % (((i*cuts) + j)*cuts + k)
                    size = float(self.temp2x[((i*cuts) + j)*cuts + k].size)
                    print "size %d" % size
                    self.densities2[((i*cuts) + j)*cuts + k], self.tags2[((i*cuts) + j)*cuts + k] = \
                        RunHOP(self.temp2x[((i*cuts) + j)*cuts + k],
                               self.temp2y[((i*cuts) + j)*cuts + k],
                               self.temp2z[((i*cuts) + j)*cuts + k],
                               self.temp2m[((i*cuts) + j)*cuts + k],
                               self.threshold,size/nParts*dens_fix)

    def __run_hop(self):
        self.densities, self.tags = \
            RunHOP(self.particle_fields["particle_position_x"],
                   self.particle_fields["particle_position_y"],
                   self.particle_fields["particle_position_z"],
                   self.particle_fields["ParticleMassMsun"],
                   self.threshold, 1.0)
        self.particle_fields["densities"] = self.densities
        self.particle_fields["tags"] = self.tags

    def __get_dm_indices(self):
        if 'creation_time' in self.data_source.hierarchy.field_list:
            mylog.debug("Differentiating based on creation time")
            return (self.data_source["creation_time"] < 0)
        elif 'particle_type' in self.data_source.hierarchy.field_list:
            mylog.debug("Differentiating based on particle type")
            return (self.data_source["particle_type"] == 1)
        else:
            mylog.warning("No particle_type, no creation_time, so not distinguishing.")
            return slice(None)

    def __parse_outputs(self):
        cuts = 3
        for i in range(cuts):
            for j in range(cuts):
                for k in range(cuts):
                    self._groups2[((i*cuts) + j)*cuts + k] = []
                    self._max_dens2[((i*cuts) + j)*cuts + k] = []
                    unique_ids = na.unique(self.tags2[((i*cuts) + j)*cuts + k])
                    counts = na.bincount(self.tags2[((i*cuts) + j)*cuts + k]+1)
                    sort_indices = na.argsort(self.tags2[((i*cuts) + j)*cuts + k])
                    grab_indices = na.indices(self.tags2[((i*cuts) + j)*cuts + k].shape).ravel()[sort_indices]
                    dens = self.densities2[((i*cuts) + j)*cuts + k][sort_indices]
                    cp = 0
                    print 'uids %d' % (unique_ids.size)
                    for ii in unique_ids:
                        cp_c = cp + counts[ii+1]
                        if ii == -1:
                            cp += counts[ii+1]
                            continue
                        group_indices = grab_indices[cp:cp_c]
                        md_i = na.argmax(dens[cp:cp_c])
                        px = self.temp2x[((i*cuts) + j)*cuts + k][group_indices]
                        py = self.temp2y[((i*cuts) + j)*cuts + k][group_indices]
                        pz = self.temp2z[((i*cuts) + j)*cuts + k][group_indices]
                        max_dens = (dens[cp:cp_c][md_i], px[md_i], py[md_i], pz[md_i])
                        # as a shortcut, if the most dense particle is in the current box, we keep the group
                        if ((math.floor(max_dens[1]*cuts) == i) and (math.floor(max_dens[2]*cuts) == j) \
                                and (math.floor(max_dens[3]*cuts) == k)):
                            # I think this will make the group_indices correct in the overall 'real' sense
                            temp_real_indices = arange(group_indices.size)
                            for gg,gi in enumerate(group_indices):
                                temp_real_indices[gg] = self.tracking2[((i*cuts) + j)*cuts + k][gi]
                            self._groups2[((i*cuts) + j)*cuts + k].append(HopGroup(self,ii, temp_real_indices))
                            self._max_dens2[((i*cuts) + j)*cuts + k].append(max_dens)
                        cp += counts[ii+1]
        for i in range(cuts**3):
            print '%d groups2 %d' % (i,len(self._groups2[i]))

    def __glue_outputs(self):
        cuts = 3
        ii = 0
        iii = 0
        for i in range(cuts):
            for j in range(cuts):
                for k in range(cuts):
                    print '%d len = %d' % (((i*cuts) + j)*cuts + k,len(self._groups2[((i*cuts) + j)*cuts + k]))
                    for group2 in self._groups2[((i*cuts) + j)*cuts + k]:
                        self._groups.append(HopGroup(self,iii,group2.indices))
                        iii += 1
                    for dens in self._max_dens2[((i*cuts) + j)*cuts + k]:
                        self._max_dens[ii] = dens
                        ii += 1

    def __parse_output(self):
        unique_ids = na.unique(self.tags)
        counts = na.bincount(self.tags+1)
        sort_indices = na.argsort(self.tags)
        grab_indices = na.indices(self.tags.shape).ravel()[sort_indices]
        dens = self.densities[sort_indices]
        cp = 0
        for i in unique_ids:
            cp_c = cp + counts[i+1]
            if i == -1:
                cp += counts[i+1]
                continue
            group_indices = grab_indices[cp:cp_c]
            self._groups.append(HopGroup(self, i, group_indices))
            md_i = na.argmax(dens[cp:cp_c])
            px, py, pz = [self.particle_fields['particle_position_%s'%ax][group_indices]
                                            for ax in 'xyz']
            self._max_dens[i] = (dens[cp:cp_c][md_i],
                                 px[md_i], py[md_i], pz[md_i])
            cp += counts[i+1]

    def __len__(self):
        return len(self._groups)
 
    def __iter__(self):
        return HopIterator(self)

    def __getitem__(self, key):
        return self._groups[key]

    def write_out(self, filename="HopAnalysis.out"):
        """
        Write out standard HOP information to *filename*.
        """
        f = open(filename,"w")
        f.write("\t".join(["# Group","Mass","# part","max dens"
                           "x","y","z", "center-of-mass",
                           "x","y","z",
                           "vx","vy","vz","max_r","\n"]))
        for group in self:
            f.write("%10i\t" % group.id)
            f.write("%0.9e\t" % group.total_mass())
            f.write("%10i\t" % group.indices.size)
            f.write("%0.9e\t" % group.maximum_density())
            f.write("\t".join(["%0.9e" % v for v in group.maximum_density_location()]))
            f.write("\t")
            f.write("\t".join(["%0.9e" % v for v in group.center_of_mass()]))
            f.write("\t")
            f.write("\t".join(["%0.9e" % v for v in group.bulk_velocity()]))
            f.write("\t")
            f.write("%0.9e\t" % group.maximum_radius())
            f.write("\n")
        f.close()

class HopIterator(object):
    def __init__(self, hop):
        self.hop = hop
        self.index = -1

    def next(self):
        self.index += 1
        if self.index == len(self.hop): raise StopIteration
        return self.hop[self.index]

class HopGroup(object):
    """
    A data source that returns particle information about the members of a
    HOP-identified halo.
    """
    def __init__(self, hop_output, id, indices):
        self.hop_output = hop_output
        self.id = id
        self.data = hop_output.data_source
        self.indices = hop_output._base_indices[indices]
        
    def center_of_mass(self):
        """
        Calculate and return the center of mass.
        """
        c_vec = self.maximum_density_location() - na.array([0.5,0.5,0.5])
        pm = self["ParticleMassMsun"]
        cx = (self["particle_position_x"] - c_vec[0])
        cy = (self["particle_position_y"] - c_vec[1])
        cz = (self["particle_position_z"] - c_vec[2])
        com = na.array([v-na.floor(v) for v in [cx,cy,cz]])
        return (com*pm).sum(axis=1)/pm.sum() + c_vec

    def maximum_density(self):
        """
        Return the HOP-identified maximum density.
        """
        return self.hop_output._max_dens[self.id][0]

    def maximum_density_location(self):
        """
        Return the location HOP identified as maximally dense.
        """
        return na.array([
                self.hop_output._max_dens[self.id][1],
                self.hop_output._max_dens[self.id][2],
                self.hop_output._max_dens[self.id][3]])

    def total_mass(self):
        """
        Returns the total mass in solar masses of the halo.
        """
        return self["ParticleMassMsun"].sum()

    def bulk_velocity(self):
        """
        Returns the mass-weighted average velocity.
        """
        pm = self["ParticleMassMsun"]
        vx = (self["particle_velocity_x"] * pm).sum()
        vy = (self["particle_velocity_y"] * pm).sum()
        vz = (self["particle_velocity_z"] * pm).sum()
        return na.array([vx,vy,vz])/pm.sum()

    def maximum_radius(self, center_of_mass=True):
        """
        Returns the maximum radius in the halo for all particles,
        either from the point of maximum density or from the (default)
        *center_of_mass*.
        """
        if center_of_mass: center = self.center_of_mass()
        else: center = self.maximum_density_location()
        rx = na.abs(self["particle_position_x"]-center[0])
        ry = na.abs(self["particle_position_y"]-center[1])
        rz = na.abs(self["particle_position_z"]-center[2])
        r = na.sqrt(na.minimum(rx, 1.0-rx)**2.0
                +   na.minimum(ry, 1.0-ry)**2.0
                +   na.minimum(rz, 1.0-rz)**2.0)
        return r.max()

    def __getitem__(self, key):
        return self.data[key][self.indices]

    def get_sphere(self, center_of_mass=True):
        """
        Returns an EnzoSphere centered on either the point of maximum density
        or the *center_of_mass*, with the maximum radius of the halo.
        """
        if center_of_mass: center = self.center_of_mass()
        else: center = self.maximum_density_location()
        radius = self.maximum_radius()
        # A bit of a long-reach here...
        sphere = self.hop_output.data_source.hierarchy.sphere(
                        center, radius=radius)
        return sphere



"""
A simple distributed object mechanism, for storing array-heavy objects.
Meant to be subclassed.

Author: Matthew Turk <matthewturk@gmail.com>
Affiliation: KIPAC/SLAC/Stanford
Homepage: http://yt.enzotools.org/
License:
  Copyright (C) 2010 Matthew Turk.  All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import numpy as na
from yt.funcs import *
from yt.parallel_tools import ParallelAnalysisInterface
from itertools import izip

class DistributedObjectCollection(ParallelAnalysisInterface):
    valid = True

    def _get_object_info(self):
        pass

    def _set_object_info(self):
        pass

    def join_lists(self):
        info_dict = self._get_object_info()
        info_dict = self._mpi_catdict(info_dict)
        self._set_object_info(info_dict)

    def _collect_objects(self, desired_indices):
        # We figure out which indices belong to which processor,
        # then we pack them up, and we send a list to each processor.
        request_count = []
        owners = self._object_owners[desired_indices]
        mylog.debug("Owner list: %s", na.unique1d(owners))
        # Even if we have a million bricks, this should not take long.
        s = self._mpi_get_size()
        m = self._mpi_get_rank()
        requests = dict( ( (i, []) for i in xrange(s) ) )
        for i, p in izip(desired_indices, owners):
            requests[p].append(i)
        for p in sorted(requests):
            requests[p] = na.array(requests[p], dtype='int64')
            request_count.append(len(requests[p]))
        size = len(request_count)
        mylog.debug("Requesting: %s", request_count)
        request_count = na.array(request_count, dtype='int64')
        # Now we distribute our requests to all the processors.
        # This is two-pass.  One to get the length of the arrays.  The second
        # pass is to get the actual indices themselves.
        request_count = self._mpi_joindict({m : request_count})
        # Now we have our final array of requests, with arrangement
        # (Nproc,Nproc).  First index corresponds to requesting proc, second to
        # sending.  So [them,us] = 5 means we owe 5, whereas [us, them] means
        # we are owed.
        send_hooks = []
        dsend_buffers, dsend_hooks = [], []
        recv_hooks, recv_buffers = [], []
        drecv_buffers, drecv_hooks = [], []
        # We post our index-list and data receives from each processor.
        mylog.debug("Posting data buffer receives")
        proc_hooks = {}
        for p, request_from in request_count.items():
            if p == m: continue
            size = request_from[m]
            #if size == 0: continue
            # We post receives of the grids we *asked* for.
            # Note that indices into this are not necessarily processor ids.
            # So we store.  This has to go before the appends or it's an
            # off-by-one.
            mylog.debug("Setting up index buffer of size %s for receive from %s",
                        size, p)
            proc_hooks[len(drecv_buffers)] = p
            drecv_buffers.append(self._create_buffer(requests[p]))
            drecv_hooks.append(self._mpi_Irecv_double(drecv_buffers[-1], p, 1))
            recv_buffers.append(na.zeros(size, dtype='int64'))
            # Our index list goes on 0, our buffer goes on 1.  We know how big
            # the index list will be, now.
            recv_hooks.append(self._mpi_Irecv_long(recv_buffers[-1], p, 0))
        # Send our index lists into hte waiting buffers
        mylog.debug("Sending index lists")
        for p, ind_list in requests.items():
            if p == m: continue
            if len(ind_list) == 0: continue
            # Now, we actually send our index lists.
            send_hooks.append(self._mpi_Isend_long(ind_list, p, 0))
        # Now we post receives for all of the data buffers.
        mylog.debug("Sending data")
        for i in self._mpi_Request_Waititer(recv_hooks):
            # We get back the index, which here is identical to the processor
            # number doing the send.  At this point, we can post our receives.
            p = proc_hooks[i]
            mylog.debug("Processing from %s", p)
            ind_list = recv_buffers[i]
            dsend_buffers.append(self._create_buffer(ind_list))
            self._pack_buffer(ind_list, dsend_buffers[-1])
            dsend_hooks.append(self._mpi_Isend_double(
                dsend_buffers[-1], p, 1))
        mylog.debug("Waiting on data receives: %s", len(drecv_hooks))
        for i in self._mpi_Request_Waititer(drecv_hooks):
            mylog.debug("Unpacking from %s", proc_hooks[i])
            # Now we have to unpack our buffers
            # Our key into this is actually the request for the processor
            # number.
            p = proc_hooks[i]
            self._unpack_buffer(requests[p], drecv_buffers[i])
        mylog.debug("Finalizing sends: %s", len(dsend_hooks))
        for i in self._mpi_Request_Waititer(dsend_hooks):
            continue

    def _create_buffer(self, ind_list):
        pass

    def _pack_buffer(self, ind_list):
        pass

    def _unpack_buffer(self, ind_list, my_buffer):
        pass




import sys
sys.path.insert(0,"/Users/matthewturk/Development/yt/trunk/")

import numpy as na
import yt.lagos as lagos

a = lagos.EnzoStaticOutput("galaxy1800.dir/galaxy1800")

sp = a.h.sphere([0.5,0.5,0.5], 1.0)

print sp["ParticleMassMsun"].max()
print sp["particle_velocity_x"].max()

print sp['particle_velocity_x'][sp['particle_type']==2].size
print sp['particle_velocity_x'][sp['particle_type']==1].size

for grid in a.h.grids:
    LE = grid.LeftEdge
    RE = grid.RightEdge
    center = (LE + RE)/2.0
    grid.set_field_parameter('center', center)
    rad = grid["RadiusCode"].max()
    reg = a.h.region(center, LE, RE)
    sph = a.h.sphere(center, rad)
    print "Checking region corresponding to grid", center, LE, RE
    # Now let's manually check our sizes
    pos = na.array([reg['particle_position_x'],
                    reg['particle_position_y'],
                    reg['particle_position_z']]).transpose()
    spos = na.array([sph['particle_position_x'],
                     sph['particle_position_y'],
                     sph['particle_position_z']]).transpose()
    mrad = na.array( na.sqrt(((spos - center)**2.0).sum(axis=1)))
    if na.any( na.all(pos > RE,axis=1) | na.all(pos < LE,axis=1) ) \
       or na.any(mrad > rad):
        print grid, na.any(mrad > rad), na.any( na.all(pos > RE,axis=1) | na.all(pos < LE,axis=1) )
        raise KeyError
